/*
 * Fixed memory segmented data storage utilising data chains.
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1.1.1 $  $Date: 2003/01/10 23:15:54 $
 * --------------------------------------------------------------------------
 */

/*
 *
 * +-------+-----+-------------+---------------------+
 * | DCtrl | Map | Chain Index | ......blocks....... |
 * +-------+-----+-------------+---------------------+
 *
 */
#include <string.h>
#include <phlib/xl_datachain.h>

#define BYTES_P_WORD 4

static unsigned int val_2_bit[] = {
  0,
  0,
  1, 0,
  2, 0,0,0,
  3, 0,0,0,0,0,0,0,
  4, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  5, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  6, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  7
};

/* prototypes */
static int  get_free_block( DCCtrl *dc );
static void release_block( int b, DCCtrl *dc );

/* ------------------------------------------------------------------------- */
/*
 * Create a datachain using the suppled memory (pointed to by dc)
 * Total memory available for storage is blocksize * numblocks
 */
void
xl_DataChain_init( DCCtrl *dc, int blocksize, int numblocks )
{
  int i;
  int mapsize  = ((numblocks / (BYTES_P_WORD * 8)) + 
		  ((numblocks % (BYTES_P_WORD * 8)) ? 1 : 0) ) * BYTES_P_WORD;

  int indexsize = numblocks * sizeof( int ) * 2; /* *2 cos of len's */

  dc->bsize   = blocksize;
  dc->o_map   = sizeof( DCCtrl );
  dc->o_index = sizeof( DCCtrl ) + mapsize;
  dc->o_data  = sizeof( DCCtrl ) + mapsize + indexsize;

  dc->mapsize   = mapsize;
  dc->freespace = blocksize * numblocks;

  /* Initialse free block map ( all 1's = all free ) */
  memset( (char*)dc + dc->o_map, 0xff, mapsize );
  
  /* Initialise index and lengths */
  for( i = 0; i < numblocks; i++ ) ((int*)((char*)dc + dc->o_index))[i] = -1;
}

/* ------------------------------------------------------------------------- */

int
xl_DataChain_mem_req( int blocksize, int numblocks )
{
  int mapsize  = ((numblocks / (BYTES_P_WORD * 8)) + 
		  ((numblocks % (BYTES_P_WORD * 8)) ? 1 : 0) ) * BYTES_P_WORD;
  int indexsize = numblocks * sizeof( int ) * 2;

  return sizeof(DCCtrl) + mapsize + indexsize + (blocksize * numblocks);
}

/* ------------------------------------------------------------------------- */
/*
 * Store data and return access handle (first block number)
 *
 * Return E_DCNOSPACE = Not Enough Space for data
 *        E_DCMEMERR  = Error allocating free block
 */
int
xl_DataChain_store( DCCtrl *dc, char* data, int len, int *blk )
{
  int nblk = -1; /* New   block */
  int lblk = -1; /* last  block */
  int fblk = -1; /* first block */
  int adj  = 0;
  int llen = len;
  char *p  = data;

  int blocks_used = 0;

#ifdef DEBUG
  printf( "Free space: %d   Inserting %d\n", dc->freespace, len );
#endif

  if( len > dc->freespace ) return E_DCNOSPACE;

  while( llen > 0 ) {
    if( (nblk = get_free_block( dc )) < 0 ) return E_DCMEMORY;

    if( fblk == -1 ) fblk = nblk; /* remember first block */

    adj = (llen > dc->bsize) ? dc->bsize : llen;
    
    if( lblk >= 0 ) ((int*)((char*)dc + dc->o_index))[ (lblk*2) + 0 ] = nblk;
    ((int*)((char*)dc + dc->o_index))[ (nblk*2) + 0 ] = -1;
    ((int*)((char*)dc + dc->o_index))[ (nblk*2) + 1 ] = adj;

    memcpy( (char*)dc + dc->o_data + (dc->bsize * nblk), p, adj );

    lblk  = nblk; 
    llen -= adj;
    p    += adj;
    blocks_used++;
  }
  dc->freespace -= ( dc->bsize * blocks_used );

  *blk = fblk;

  return E_DCNOERROR;
}


/* ------------------------------------------------------------------------- */
/*
 * Append data to an existing chain.
 *
 * Entry: Ptr to DC
 *        Ptr to data to append
 *        Length of data to append
 *        Last block number of chain to append to
 *
 * Exit : The new Last block number of chain.
 */
int
xl_DataChain_append( DCCtrl *dc, char *data, int len, int *olblk )
{
  int llen = len;
  int nblk = -1; /* New   block */
  int lblk = -1; /* last  block */
  int adj  = 0;
  int olen = 0;
  char *p  = data;

  int blocks_used = 0;

  if( len > dc->freespace ) return E_DCNOSPACE;

  nblk = lblk = *olblk;

  /*
   * The last block of the existing chain may not be fully populated.
   * If it's not, then we need to fill it first
   */
  olen = ((int*)((char*)dc + dc->o_index))[ (lblk*2) + 1 ]; /* Len lst blk */

  if( olen < dc->bsize ) {

    adj = (dc->bsize - olen >= len) ? len : dc->bsize - olen;

    memcpy( (char*)dc + dc->o_data + (dc->bsize * lblk) + olen, p, adj );
    p    += adj;
    llen -= adj;
  }

  /* If we've any more data left, we need to append new blocks to the chain
   */
  while( llen > 0 ) {
    if( (nblk = get_free_block( dc )) < 0 ) return E_DCMEMORY;

    adj = (llen > dc->bsize) ? dc->bsize : llen;
    
    ((int*)((char*)dc + dc->o_index))[ (lblk*2) + 0 ] = nblk;

    ((int*)((char*)dc + dc->o_index))[ (nblk*2) + 0 ] = -1;
    ((int*)((char*)dc + dc->o_index))[ (nblk*2) + 1 ] = adj;

    memcpy( (char*)dc + dc->o_data + (dc->bsize * nblk), p, adj );

    lblk  = nblk; 
    llen -= adj;
    p    += adj;
    blocks_used++;
  }
  dc->freespace -= ( dc->bsize * blocks_used );

  *olblk = nblk;

  return E_DCNOERROR;
}



/* ------------------------------------------------------------------------- */
/*
 * Returns data a block at a time.
 *
 * Entry : Ptr to DC
 *         Block number to retrieve
 *
 * Exit  : Ptr to data block
 *         Length of data in block
 * 
 * Return: Block number of next block in chain (or -1 if there are no more)
 */
int
xl_DataChain_get_data( DCCtrl *dc, int block, char **dptr, int *len )
{
  *dptr = (char*)dc + dc->o_data + (dc->bsize * block);
  *len  = ((int*)((char*)dc + dc->o_index))[ (block * 2) + 1 ];

  return ((int*)((char*)dc + dc->o_index))[ (block * 2) + 0 ];
}

/* ------------------------------------------------------------------------- */

int
xl_DataChain_free( DCCtrl *dc, int block )
{
  int len  = 0;
  int nblk = block;
  int bused = 0;

  do {
    release_block( nblk, dc );

    len  += ((int*)((char*)dc + dc->o_index))[ (nblk * 2) + 1 ];
    nblk = ((int*)((char*)dc + dc->o_index))[ (nblk * 2) + 0 ];
    bused++;
  } while( nblk >= 0 );

#ifdef DEBUG
  printf( "Increase Free space: %d   by %d\n", dc->freespace, dc->bsize * bused );
#endif
  dc->freespace += ( dc->bsize * bused );

  return dc->freespace;
}

/* ------------------------------------------------------------------------- */
/*
 * This function MUST BE USED WITH CARE!
 *
 * It will return the block number of the next block in the chain.
 */
int 
xl_DataChain_free_block( DCCtrl *dc, int block )
{
  int nextblock  = 0;
  int len = 0;

  if( block < 0 )  return -1;

  release_block( block, dc );

  len       = ((int*)((char*)dc + dc->o_index))[ (block * 2) + 1 ];
  nextblock = ((int*)((char*)dc + dc->o_index))[ (block * 2) + 0 ];

#ifdef DEBUG
  printf( "Increase Free space: %d   by %d\n", dc->freespace, dc->bsize );
#endif

  dc->freespace += dc->bsize;

  return nextblock;
}

/* ------------------------------------------------------------------------- */

/*
 * This PRIVATE function locates a free 'block' in our storage area
 * and marks it as being used.  The block number is returned.
 */
static int
get_free_block( DCCtrl *dc )
{
  RetValue ret = Failure;
  unsigned int *map;
  unsigned int i, b, w, t;

  map = (unsigned int*)( (char*)dc + dc->o_map );
  i = dc->mapsize * 8;

  b = 0;
  while( b < i ) {
    w = *map;
    t = w & -w;
    if( t ) {
      /* Clear bit, indicating that it's now being used */
      *map &= ~t;
  
      /* work out the bit number (block number) and return it */
      if     ( t > 0x00ffffff ) b += val_2_bit[ t >> 24 ] + 24;
      else if( t > 0x0000ffff ) b += val_2_bit[ t >> 16 ] + 16;
      else if( t > 0x000000ff ) b += val_2_bit[ t >> 8  ] + 8;
      else b += val_2_bit[ t ];
      ret = Success;
      break;
    }
    b += 32;
    map++;
  }
  
  return ret == Success ? b : -1;
}


/* ------------------------------------------------------------------------- */
/*
 * Set a bit in the block map indicating that it is free
 */
static void
release_block( int b, DCCtrl *dc )
{
  ((int*)( (char*)dc + dc->o_map ))[ b/32 ] |= 1 << (b%32);
}


/* ------------------------------------------------------------------------- */
/* end */
