/*
 * Linked List object
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1.1.1 $  $Date: 2003/01/10 23:16:00 $
 * --------------------------------------------------------------------------- 
 */

/*
 * Notes:
 * -----
 * The memory malloced for a node is sizeof ll_node + sizeof data struct
 * (the data struct is user supplied, and ptr of ll_node part point to it...
 *
 * Illustration:
 * ...............................
 * | +-> next ------>----------+
 * +-|-< prev                  |
 *   |   ptr ------->-------+  |
 *   |     applic struct <--+  |
 *   |                         |
 *   |    +---------<----------+
 * ..|....|.......................
 * +-|-> next ------>----------+
 * | +-< prev                  |
 * |     ptr ------->--------+ |
 * |        applic struct <--+ |
 * |                           |
 * |      +---------<----------+
 * |......|.......................
 * | +-> next ------>----------+
 * +-|-< prev                  |
 *   |   ptr ------->-------+  |
 *   |     applic struct <--+  |
 *   |                         |
 *   |    +---------<----------+
 * ..|....|.......................
 */

#include <stdlib.h>
#include <string.h>

#include <phlib/xl_config.h>
#include <phlib/xl_types.h>
#include <phlib/xl_linklst.h>
#include <phlib/xl_malloc.h>

/* Private prototypes */
static RetValue xlp_create_node ( Node **node, LIST *listref );
static RetValue xlp_detatch_node( Node **node, LIST *listref );
/* ---------------------------------------------------------------------- */
/*
 * Initialise a linked list.
 * You pass this function details of the data segment you wish to store in
 * this list.  The function returns a 'handle' which is used to identify the
 * list you are interested in in all subsequent list operations.
 */
RetValue
xl_List_init( LIST **listref, int user_data_size )
{
  LIST *newlist;
  RetValue ret = Failure;

  if( (newlist = (LIST*)xl_malloc( sizeof( LIST ) )) != NULL ) {
    newlist->count    = 0;
    newlist->datasize = user_data_size;
    newlist->nodesize = user_data_size + sizeof( Node );
    newlist->head     = NULL;
    newlist->last     = NULL;
    newlist->current  = NULL;

    *listref = newlist;

    ret = Success;
  }
  return ret;
}


/* ---------------------------------------------------------------------- */
/*
 * Free up a linked list
 */ 
RetValue
xl_List_end( LIST **listref )
{
  RetValue ret = Failure;
  Node *ptr, *kill;

  if( *listref != NULL ) {
    ptr = (*listref)->head;

#ifdef DEVELOPMENT
    printf( "Freeing linked list....\n" );    
#endif
    /* Free all list nodes */
    while( ptr != NULL  ) {
      kill = ptr;
      ptr = ptr->next;
      xl_free( (uchar_t*)kill );
    }

    /* Free list identifier */
    xl_free( (uchar_t*)(*listref) );
    *listref = NULL;
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */
/*
 * THIS IS CALLED VIA MACROS - xl_List_insert_node and ...insert_space
 *
 * This function inserts nodes into the linked list.
 * It will insert at the Start, End or somewhere in the middle.....
 * If the insert position indicator is greater then the number of nodes
 * in the list then a Failure return will occur.
 */
RetValue
xl_List_insert_a_node( LIST *listref, ListPos pos, uchar_t *userdata, int set )
{
  Node *new, *keep;
  RetValue ret = Failure;
  ListPos  mpos = pos;
  int i;

  ASSERT( listref );

  /* Fixup incase list is empty or we're on the last node */
  if( mpos == listref->count ) mpos = ListEnd;
  if( listref->count == 0 )    mpos = 0;

  switch( mpos ) {
  case 0:
  case ListStart: 
    if( xlp_create_node( &new, listref ) == Success ) {
      keep = listref->head;
      listref->head = new;
      if(keep) keep->prev = new;
      new->next     = keep;
      new->prev     = NULL;
      mpos          = 0;
      ret = Success;
    }
    break;

  case ListEnd:
    if( xlp_create_node( &new, listref ) == Success ) {
      listref->last->next = new;
      new->prev           = listref->last;
      new->next           = NULL;
      listref->last       = new; /* We know we're the new end tag.... */
      ret = Success;
    }
    break;

  default:
    if( mpos <= listref->count && mpos > 0 ) {
      if( xlp_create_node( &new, listref ) == Success ) {
	
	keep = listref->head; /* Point at first node */
	for( i = 0; i < mpos; i++ ) keep = keep->next; /* Find require node */
	new->prev        = keep->prev;
	keep->prev->next = new;
	new->next        = keep;
	keep->prev       = new;
	ret = Success;
      }
    }
    break;
  }

  if( ret == Success ) {
    if( mpos == listref->count ) listref->last = new; /* If new end tag */
    listref->count++;
    listref->current = new;            /* Record last node accessed */
    new->nodesize = listref->nodesize; /* Record size of node       */

    /* Insert user data into structure if set is 1 */
    if(set == 1) memcpy( new->ptr, userdata, listref->datasize );
  }
  return ret;
}

/* ---------------------------------------------------------------------- */
/*
 * This function detatches the current node from one list and inserts it
 * in a second list at the desired position
 */
RetValue
xl_List_transfer_node( LIST *listref, LIST *src, ListPos pos )
{
  Node *new, *keep;
  RetValue ret = Failure;
  ListPos  mpos = pos;
  int i;

  ASSERT( listref );
  ASSERT( src     );

  /* Fixup incase list is empty or we're on the last node */
  if( mpos == listref->count ) mpos = ListEnd;
  if( listref->count == 0 )    mpos = 0;

  switch( mpos ) {
  case 0:
  case ListStart: 
    if( xlp_detatch_node( &new, src ) == Success ) {
      keep = listref->head;
      listref->head = new;
      if(keep) keep->prev = new;
      new->next     = keep;
      new->prev     = NULL;
      mpos          = 0;
      ret = Success;
    }
    break;
    
  case ListEnd:
    if( xlp_detatch_node( &new, src ) == Success ) {
      listref->last->next = new;
      new->prev           = listref->last;
      new->next           = NULL;
      listref->last       = new; /* We know we're the new end tag.... */
      ret = Success;
    }
    break;

  default:
    if( mpos <= listref->count && mpos > 0 ) {
      if( xlp_detatch_node( &new, src ) == Success ) {
	
	keep = listref->head; /* Point at first node */
	for( i = 0; i < mpos; i++ ) keep = keep->next; /* Find require node */
	new->prev        = keep->prev;
	keep->prev->next = new;
	new->next        = keep;
	keep->prev       = new;
	ret = Success;
      }
    }
    break;
  }

  if( ret == Success ) {
    if( mpos == listref->count ) listref->last = new; /* If new end tag */
    listref->count++;
    listref->current = new;            /* Record last node accessed */
  }
  return ret;
}


/* ---------------------------------------------------------------------- */

char *
xl_List_get_current_data( LIST *listref )
{
  return (char*)(listref->current->ptr);
}

/* ---------------------------------------------------------------------- */

int
xl_List_get_size( LIST *listref )
{
  ASSERT( listref );
  return listref->count;
}

/* ---------------------------------------------------------------------- */
/*
 * This function allows code to re-size the userdata portion of a node
 * prior to insertion of a new node.  This provides for nodes which must
 * contain variable sized data portions.  
 * It is the applications responsibility to know the datasize of a particular
 * node when reading it.  This sets the size for ALL susequent inserts.
 * Entry: List descriptor, new datasize
 * Exit : Old datasize
 */
int
xl_List_set_datasize( LIST *listref, int user_data_size )
{
  int oldsize;

  ASSERT( listref );

  oldsize = listref->datasize;

  listref->datasize = user_data_size;
  listref->nodesize = user_data_size + sizeof( Node );  

  return oldsize;
}

/* ---------------------------------------------------------------------- */
/*
 * This function inserts nodes into the linked list - But with an additional
 * parameter setting the user datasize for THIS and ONLY THIS node.
 * It will insert at the Start, End or somewhere in the middle.....
 * If the insert position indicator is greater then the number of nodes
 * in the list then a Failure return will occur.
 */
RetValue
xl_List_insert_nodeS( LIST *listref, ListPos pos, uchar_t *userdata, int uds )
{
  int oldsize;
  RetValue ret;

  ASSERT( listref );

  oldsize = listref->datasize;

  listref->datasize = uds;
  listref->nodesize = uds + sizeof( Node );  

  ret = xl_List_insert_node( listref, pos, userdata );

  listref->datasize = oldsize;
  listref->nodesize = oldsize + sizeof( Node );  

  return ret;
}
 
/* ---------------------------------------------------------------------- */

int
xl_List_get_nodesize( LIST *listref )
{
  Node *cur;
  ASSERT( listref );

  cur  = listref->current;

  if( cur == NULL ) return (-1);
  else return cur->nodesize;
}

/* ---------------------------------------------------------------------- */

RetValue
xl_List_get_node( LIST *listref, ListPos pos, uchar_t **userdata )
{
  RetValue ret = Failure;
  Node *ptr;
  int i;

  ASSERT( listref );

  if( pos < listref->count ) {
    ptr = listref->head; /* Point at first node */
    for( i = 0; i < pos; i++ ) ptr = ptr->next; /* Find require node */

    listref->current = ptr;     /* Last node accessed */
    *userdata = ptr->ptr;
    ret = Success;
  }
  return ret;
}

/* ---------------------------------------------------------------------- */
/*
 * This deletes the current list node.
 * If the node is in the middle of the list (ie not first or last) then
 * the new 'current' node becomes the one on the left.
 * However, if we are deleteing the first node, then the new first node
 * becomes the node to the right and also the current node!
 */
RetValue
xl_List_delete_node( LIST *listref )
{
  RetValue ret = Failure;
  Node *right, *kill, *left;

  ASSERT( listref );
  kill = listref->current;

  if( kill != NULL ) {
    right = kill->next;
    left  = kill->prev;
    xl_free( (uchar_t*)kill ); /* Free memory block */
    
    /* If there is no right node then we're deleting the last node */
    if( right == NULL ) {
      listref->last = left; /* New last node */
    } else {
      right->prev = left;
    }
    
    if( left  != NULL ) {
      left->next       = right;
      listref->current = left;  /* The left one is the last accessed.. */
    } else {
      /* Hmm, there is no left node! */
      listref->head    = right;
      listref->current = right; /* Of course.. this may be NULL!       */
    }

    listref->count--;
    
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */

RetValue
xl_List_get_first( LIST *listref, uchar_t **userdata )
{
  RetValue ret = Failure;
  Node *ptr;

  ASSERT( listref );
  ptr = listref->head;          /* Point to first cell */
  if( ptr != NULL ) {
    listref->current = ptr;     /* Last node accessed */
    *userdata = ptr->ptr;
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */
/*
 * Moves the current node to the head of the list 
 * Current node remains the current node! and hence the NEXT node is the
 * _old_ head of the list.
 */
RetValue
xl_List_make_first( LIST *listref )
{
  Node *cur;
  ASSERT( listref );
  cur  = listref->current;

  /* Check we're not the first node in the list! */
  if( cur != NULL && cur->prev != NULL ) { 
    cur->prev->next = cur->next;

    if(cur->next) { cur->next->prev = cur->prev; /* Join next to prev */
    } else { listref->last            = cur->prev; /* New last node */
    }

    cur->prev       = NULL;
    cur->next       = listref->head;
    listref->head   = cur;
    cur->next->prev = cur;
  } 

  return Success;
}

/* ---------------------------------------------------------------------- */

RetValue
xl_List_get_last( LIST *listref, uchar_t **userdata )
{
  RetValue ret = Failure;
  Node *ptr;

  ASSERT( listref );
  ptr = listref->last;          /* Point to last cell */
  if( ptr != NULL ) {
    listref->current = ptr;     /* Last node accessed */
    *userdata = ptr->ptr;
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */

RetValue
xl_List_get_next( LIST *listref, uchar_t **userdata )
{
  RetValue ret = Failure;
  Node *ptr;

  ASSERT( listref );
  ptr = listref->current;       /* Point to last node accessed */

  if( ptr != NULL && ptr->next != NULL ) {

    ptr = ptr->next;            /* Step forward to next node   */
    listref->current = ptr;     /* Remember last node accessed */
    *userdata = ptr->ptr;
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */

RetValue
xl_List_get_prev( LIST *listref, uchar_t **userdata )
{
  RetValue ret = Failure;
  Node *ptr;

  ASSERT( listref );
  ptr = listref->current;       /* Point to last node accessed */

  if( ptr != NULL && ptr->prev != NULL ) {

    ptr = ptr->prev;            /* Step forward to next node   */
    listref->current = ptr;     /* Remember last node accessed */
    *userdata = ptr->ptr;

    ret = Success;
  }

  return ret;
}
/* ---------------------------------------------------------------------- */
/*
 * This PRIVATE function creates a new but 'free' list node.
 * It is the responsibillity of the caller to 'attach' the node to the list
 */
static RetValue
xlp_create_node( Node **node, LIST *listref )
{
  RetValue  ret;
  Node     *ptr;

  ASSERT( listref );
  ret  = Failure;
  *node = NULL;

  if( (ptr = (Node *)xl_malloc( listref->nodesize )) != NULL ) {
    /* Wrap list node around new memory area and point to user segment */
    *node = ptr;
    /* Set internal ptr offset to start of user data area for node     */
    ptr->ptr = (uchar_t*)ptr + sizeof( Node );
    ret = Success;
  }
  return ret;
}

/* ---------------------------------------------------------------------- */
/*
 * Detatch the current node from the list and return a pointer to it
 */
static RetValue
xlp_detatch_node( Node **node, LIST *listref )
{
  RetValue ret = Failure;
  Node *right, *kill, *left;

  ASSERT( listref );
  kill = listref->current;

  if( kill != NULL ) {
    right = kill->next;
    left  = kill->prev;

    /* If there is no right node then we're detatching the last node */
    if( right == NULL ) {
      listref->last = left; /* New last node */
    } else {
      right->prev = left;
    }
    
    if( left  != NULL ) {
      left->next       = right;
      listref->current = left;  /* The left one is the last accessed.. */
    } else {
      /* Hmm, there is no left node! */
      listref->head    = right;
      listref->current = right; /* Of course.. this may be NULL!       */
    }

    listref->count--;
    
    ret = Success;
  }

  *node = kill;

  return ret;
}

/* ---------------------------------------------------------------------- */
/* end */
