/*
 * File Mapping functions
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/04/13 22:43:37 $
 *
 * Created: 13 Dec 1996
 *
 * ---------------------------------------------------------------------------
 */

#include <phlib/xl_config.h>
#include <phlib/xl_types.h>
#include <phlib/xl_mmap.h>

#define RETRIES 10  /* 10 Retries spanning 10 seconds to aquire lock */

#ifndef O_SYNC
#  define O_SYNC 0
#endif

#if HAVE_MMAP
/* ----------------------------------------------------------------------- */
/*
 * This returns 0 if mapped, < 0 in error.. see -ve ret codes below 
 */
int
xl_mapfile( const char *fname, CString *buffCS )
{
  caddr_t mem;
  struct  stat sb;
  int     fd = MINUSONE;
  int     ret = 0;
      
  if( (fd = open( (char*)fname, O_RDONLY ) ) < 0 ) ret = (-3);
  if( ret == 0 && fstat( fd, &sb ) < 0 ) ret = (-2);
  
  if( ret == 0 ) {
    if( (mem = mmap( (caddr_t)0, (int)sb.st_size, 
		     PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0 )) >= (caddr_t)0 ) {
      WrapCStringPtrLen( buffCS, (char*)mem, (int)sb.st_size );
    } else ret = (-1);
  }

  if( ret > (-3) ) close( fd );

  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * Map a file and return a pointer to it's stat details
 * This returns 0 if mapped, < 0 in error.. see -ve ret codes below 
 */
int
xl_mapfile_stat( const char *fname, CString *buffCS, struct stat **sp )
{
  caddr_t mem;
  static struct  stat sb;
  int     fd = MINUSONE;
  int     ret = 0;
      
  if( (fd = open( (char*)fname, O_RDONLY ) ) < 0 ) ret = (-3);
  if( ret == 0 && fstat( fd, &sb ) < 0 ) ret = (-2);
  
  if( ret == 0 ) {
    if( (mem = mmap( (caddr_t)0, (int)sb.st_size, 
		     PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0 )) >= (caddr_t)0 ) {
      WrapCStringPtrLen( buffCS, (char*)mem, (int)sb.st_size );
    } else ret = (-1);
  }

  if( ret > (-3) ) close( fd );
  *sp = &sb;

  return ret;
}
/* ----------------------------------------------------------------------- */

RetValue
xl_unmapfile( CString *bufCS )
{
  RetValue ret = Failure;
  if( bufCS->p != 0 && bufCS->max != 0 ) {
    munmap( (caddr_t)(bufCS->p), (size_t)(bufCS->max) );
    ret = Success;
  }
  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * This returns the fd of the mapped file LOCKED! in success, or < 0 in error
 * The file must be unlocked and closed when done (xl_unmapfile_lock does
 * this anyway).
 */
int
xl_mapfile_lock( const char *fname, CString *buffCS  )
{
  caddr_t mem;
  struct  stat sb;
  int     i = 0;
  int     ret = 0;

  int     fd = MINUSONE;

  /* Do a few retries to get hold of the file */      
  if((fd = open((char*)fname, O_RDWR | O_SYNC, 0666)) < 0 && i++ < RETRIES ) sleep(1);
  if( i >=  RETRIES ) ret = (-3);

  if( ret == 0 && xl_flock( fd ) == Failure ) ret = (-4);
  if( ret == 0 && fstat( fd, &sb ) < 0 ) ret = (-2);
  
  if( ret == 0 ) {
    if( (mem = mmap( (caddr_t)0, (int)sb.st_size, 
		     PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0 )) >= (caddr_t)0 ) {
      WrapCStringPtrLen( buffCS, (char*)mem, (int)sb.st_size );
      ret = fd;
    } else ret = (-1);
  }

  if( ret < 0 && ret != (-3) ) { /* Hmm, error */
    xl_funlock( fd );
    close( fd );
  }

  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
xl_unmapfile_lock( CString *bufCS, int fd )
{
  RetValue ret = Failure;

  if( bufCS->p != 0 && bufCS->max != 0 ) {
    munmap( (caddr_t)(bufCS->p), (size_t)(bufCS->max) );
    xl_funlock( fd );
    close( fd );
    ret = Success;
  }

  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * This returns 0 if mapped, < 0 in error.. see -ve ret codes below 
 * IT DOES NOT CLOSE THE FILE
 */
int
xl_mapfile_fd( int fd, CString *buffCS )
{
  caddr_t mem;
  struct  stat sb;
  int     ret = 0;
      
  if( ret == 0 && fstat( fd, &sb ) < 0 ) ret = (-2);
  
  if( ret == 0 ) {
    if( (mem = mmap( (caddr_t)0, (int)sb.st_size, 
		     PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0 )) >= (caddr_t)0 ) {
      WrapCStringPtrLen( buffCS, (char*)mem, (int)sb.st_size );
    } else ret = (-1);
  }

  return ret;
}
#endif /* HAVE_MMAP */

/* ----------------------------------------------------------------------- */
/* end */
