/*
 * ------------------------------------------------------------------------
 * Module : TCP/IP Socket Library
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/04/13 19:53:49 $
 *
 * Created: 11 Nov 1996
 * ---------------------------------------------------------------------------
 */

#include <phlib/xl_config.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>

#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif

#ifdef MACH_SunOS
#  include <strings.h>
#endif

#if STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

#include <phlib/xl_types.h>
#include <phlib/xl_socket.h>
#include <phlib/xl_ipaddr.h>

#ifndef HAVE_GETHOSTNAME
struct hostent *gethostname( const char *, int );
#endif

static char HostName[256] = "\0";

/* ---------------------------------------------------------------------- */

int
xl_get_connection( const char *host, int port )
{
  struct sockaddr_in  sin;
  struct hostent     *hp;
  int                 sock;
  unsigned long       inaddr;
  
  /*
   * Create Address to Bind to
   */
  inaddr = inet_addr( host );

  if( (long)inaddr != (-1) ) /* Already in nnn.nnn.nnn.nnn format */
    memcpy( &sin.sin_addr, &inaddr, sizeof(inaddr) );
  else {           /* Convert jobbie.co.uk into numeric nnn.nnn.nnn.nnn */
    hp = gethostbyname( host );
    if( hp == NULL ) return (-2);
    memcpy( &sin.sin_addr, hp->h_addr, hp->h_length );
  }
  
  sin.sin_family = AF_INET;        /* Unix domain                   */
  sin.sin_port   = htons( port );  /* Convert port to network order */

  /* 
   * Socket...
   */
  sock = socket( AF_INET, SOCK_STREAM, 0 );

  if( sock >= 0 &&
      connect( sock, (struct sockaddr *)&sin, 
	       sizeof(sin)) < 0 ) sock = (-3);    
  return sock;
}

/* ---------------------------------------------------------------------- */

int
xl_bind_port( int port )
{
  struct sockaddr_in  sin;
  struct hostent     *hp;
  int                 sock, opt;

  /*
   * Create Address to Bind to
   */
  if( *HostName == '\0' ) {
    gethostname( HostName, sizeof( HostName ) );
  }

  if ((hp = gethostbyname( HostName )) == NULL ) {
    fprintf( stderr, "%s: host unknown.\n", HostName );
    return MINUSONE;
  }

  /* 
   * Socket...
   */
  sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );

  /* Set socket up for immediate reuse when dropped */
  opt = 1;
  if( setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, 
	 	  (char*)&opt, sizeof(int)) < 0 ) sock = MINUSONE;

  bcopy( hp->h_addr, &sin.sin_addr, hp->h_length );
  sin.sin_family = AF_INET;        /* Unix domain                       */
  sin.sin_port   = htons( port );  /* Convert port to network order     */

  if( sock >= 0 &&
      bind( sock, (struct sockaddr *)&sin, 
	    sizeof(sin) ) < 0 ) sock = MINUSONE;

  return sock;
}

/* ---------------------------------------------------------------------- */

int
xl_bind_port_IP( int port, const char *host )
{
  struct sockaddr_in  sin;
  struct hostent     *hp;
  int                 sock, opt;
  unsigned long       inaddr;
  
  /*
   * Create Address to Bind to
   */
  inaddr = inet_addr( host );

  if( (long)inaddr != (-1) ) /* Already in nnn.nnn.nnn.nnn format */
    bcopy( &inaddr, &sin.sin_addr, sizeof(inaddr) );
  else {           /* Convert jobbie.co.uk into numeric nnn.nnn.nnn.nnn */
    if( (hp = gethostbyname( host )) == NULL ) {
      fprintf( stderr, "%s: host unknown.\n", HostName );
      return MINUSONE;
    }
    bcopy( hp->h_addr, &sin.sin_addr, hp->h_length );
  }
  
  sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );

  /* Set socket up for immediate reuse when dropped */
  opt = 1;
  if( setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, 
	 	  (char*)&opt, sizeof(int)) < 0 ) sock = MINUSONE;

  sin.sin_family = AF_INET;        /* INET domain                       */
  sin.sin_port   = htons( port );  /* Convert port to network order     */
  
  if( sock >= 0 &&
      bind( sock, (struct sockaddr *)&sin, 
	    sizeof(sin) ) < 0 ) sock = MINUSONE;
  
  return sock;
}

/* ---------------------------------------------------------------------- */
/* THIS DOES NOTHING!!! Was in the UNIX DOMAIN AND DID AN UNLINK! */
int
xl_free_bind( int port )
{
  struct sockaddr_in  sin;
  struct hostent     *hp;

  /*
   * Create Address to Bind to
   */
  if( *HostName == '\0' ) {
    gethostname( HostName, sizeof( HostName ) );
  }
  if ((hp = gethostbyname( HostName )) == NULL ) {
    fprintf( stderr, "%s: host unknown.\n", HostName );
    return MINUSONE;
  }

  bcopy( hp->h_addr, &sin.sin_addr, hp->h_length );
  sin.sin_family = AF_INET;        /* Unix domain                       */
  sin.sin_port   = htons( port );  /* Convert port to network order     */

  return 0;
}


/* ---------------------------------------------------------------------- */
/* 
 * Converts the string representation of the ipaddress into a numeric string
 * representation.
 *
 * ie imagepage.co.uk -> aaa.bbb.ccc.ddd
 *    10.1.1.2        -> 10.1.1.2
 */
char *
xl_ipaddr_numeric( char *addr )
{
  struct in_addr  inp  = {0};
  struct hostent *hp   = NULL;
  static char naddr[32] = {0};

  /* Convert Sched hostname into aa.bb.cc.dd.ee form
   */
  if( inet_aton( addr, &inp ) != 0 ) return addr; /* Fine as it is */

  /* Address not in a.b.c.d format.  Convert it. */
  if( (hp = gethostbyname( addr )) == NULL )  return NULL;

  xl_get_IPAddr_str( naddr, *((ulong*)hp->h_addr) );

  return naddr;
}

/* ---------------------------------------------------------------------- */

int 
xl_sock_disable_nagle( int sock ) 
{
  int opt = 0;
  int ret = 0;

  opt = 1;
  if( setsockopt( sock, IPPROTO_TCP, TCP_NODELAY, 
	 	  (char*)&opt, sizeof(int)) < 0 ) ret = MINUSONE;

   return ret;
}

/* ---------------------------------------------------------------------- */
/* end */
